"""Wherein functionality concerning listing basket jsons is contained."""
import os

from fsspec import AbstractFileSystem

def _get_list_of_basket_jsons(
    root_dir: str, file_system: AbstractFileSystem
) -> list[str]:
    """Return a list of basket manifest paths in the given root dir.

    Parameters:
    -----------
    root_dir: str
        Path to search for basket manifests--doesn't have to be the pantry root
    file_system: fsspec object
        The file system to search in.

    Returns:
    ----------
    A list of paths to basket_manifest.json files found under the given dir
    """
    # In some instances, root_dir may be the empty str and should be left as is
    root_dir = os.path.normpath(root_dir) if root_dir != '' else ''
    # On Windows find() returns with forward slashes, the root dir must match.
    root_dir = root_dir.replace(os.sep, '/')
    manifest_paths = []
    for path in file_system.find(root_dir):
        if path.endswith("basket_manifest.json"):
            # The find method returns absolute paths which need to be trimmed
            # to start from the root_dir instead of the full path.
            # This is done to ensure returned paths are relative to the pantry.
            manifest_paths.append(path[path.index(root_dir):])
    return manifest_paths



"""Contains scripts concerning Mongo Loader functionality."""

# Ignore pylint duplicate code. Code here is used to explicitly show pymongo is
# an optional dependency. Duplicate code is found in config.py (where pymongo
# is actually imported)
# pylint: disable-next=duplicate-code
try:
    import pymongo
except ImportError:
    _HAS_PYMONGO = False
else:
    _HAS_PYMONGO = True

from .config import get_mongo_db

# pylint: disable-next=too-many-instance-attributes
class MongoLoader():
    """Initializes a mongo loader class. Retrieves a connection to the mongo
    server and facilitates the uploading of records to each pantry's individual
    mongo db based on the record type (ie supplement, manifest, metadata).
    """

    def __init__(self, pantry, mongo_client=None, **kwargs):
        """Creates the mongo loader and makes a reference to the pantry's DB.

        Parameters
        ----------
        pantry: weave.Pantry
            A pantry object
        mongo_client: pymongo.MongoClient or None (optional; defaults to None)
            A pre-constructed MongoClient to be used. Ignored if the Pantry arg
            has a non-None pantry.mongo_client. Otherwise, this arg will be
            used for the client. If none, is provided, a MongoClient will be
            constructed using the optional mongo_config dictionary,
            or weave.config.get_mongo_db() as a last resort.
        **mongo_config: dict (optional)
            Dictionary containing the configuration settings of this loader.
            Supported Keys:
        """
        if not _HAS_PYMONGO:
            raise ImportError("Missing Dependency. The package 'pymongo' "
                              "is required to use this class.")

        self.pantry = pantry
        self.mongo_config = kwargs.get("mongo_config", {})

        # Priorize using the pantry client, otherwise use the given client.
        self.mongo_client = self.pantry.mongo_client or mongo_client
        # If both were None, try to make a client using the mongo_config.
        if self.mongo_client is None and self.mongo_config:
            self.mongo_client = pymongo.MongoClient(
                host=self.mongo_config["mongodb_host"],
                username=self.mongo_config["mongodb_username"],
                password=self.mongo_config["mongodb_password"],
                port=self.mongo_config.get("mongodb_port", 27017),
                timeoutMS=self.mongo_config.get("mongodb_timeout", None)
            )
            self.mongo_client.server_info()
        # If we still don't have a valid mongo_client, use the weave config as
        # a last resort.
        if self.mongo_client is None:
            timeout = self.mongo_config.get("mongodb_timeout", None)
            self.mongo_client = get_mongo_db(timeout=timeout)

        # Get the database. (Use MONGODB_DATABASE, defaulting to
        # pantry_path if it is not present.)
        self.database_name = self.mongo_config.get(
            "mongodb_database", self.pantry.pantry_path)
        for invalid_char in [" ", ".", "$", "/", "\\", "\x00", '"']:
            if invalid_char in self.database_name:
                self.database_name = self.database_name.replace(
                    invalid_char,
                    "",
                )
        self.database = self.mongo_client[self.database_name]

        self.metadata_collection = self.mongo_config.get(
            "metadata_collection", "metadata")
        self.manifest_collection = self.mongo_config.get(
            "manifest_collection", "manifest")
        self.supplement_collection = self.mongo_config.get(
            "supplement_collection", "supplement")


    def load_mongo_metadata(self, uuids: list[str], **kwargs):
        """Load metadata from baskets into the mongo database.

        A metadata.json is created in baskets when the metadata
        field is provided upon upload. This metadata is added to the
        Mongo database when invoking load_mongo_metadata. UUID and basket_type
        from the index_table are also added to Mongo for referrence
        back to the datasource.

        Parameters
        ----------
        uuids: [str]
            A list of uuids to add their metadata to the mongo db.
        **collection: str (optional; defaults to self.metadata_collection)
            Metadata will be added to the Mongo collection specified. If not
            provided, populate using self.metadata_collection (which defaults
            to "metadata" if otherwise unspecified).
        **replace: bool (optional; defaults to False)
            If True, the metadata for the given uuids will be replaced in the
            MongoDB collection. If False, the metadata will only be added if
            it does not already exist in the MongoDB collection.
        **metadata_dict: dict (optional)
            A dictionary containing metadata for a single basket uuid to be
            added to the MongoDB. This offers a way to add metadata that
            perserves the original types of the metadata fields
            (e.g. datetime, int, float, etc.). If provided, only one uuid
            is allowed in the uuids list. If not provided, the metadata will be
            retrieved from the basket's metadata.json file.
        """
        collection = kwargs.get("collection", self.metadata_collection)
        if not isinstance(uuids, list):
            uuids = [uuids]
        if not isinstance(uuids[0], str):
            raise TypeError("Invalid datatype for uuids: "
                            "must be a list of strings [str]")
        if not isinstance(collection, str):
            raise TypeError("Invalid datatype for metadata collection: "
                            "must be a string")

        metadata_dict = kwargs.get("metadata_dict", None)
        if metadata_dict and len(uuids) != 1:
            raise ValueError("If metadata_dict is provided only one uuid is "
            "allowed.")

        for uuid in uuids:
            basket = self.pantry.get_basket(uuid)
            if metadata_dict:
                metadata = metadata_dict
            else:
                metadata = basket.get_metadata()
            if not metadata:
                continue
            mongo_metadata = {}
            mongo_metadata["uuid"] = basket.uuid
            mongo_metadata["basket_type"] = basket.basket_type
            mongo_metadata["parent_uuids"] = basket.parent_uuids
            mongo_metadata.update(metadata) # Prioritize the metadata provided

            # If the UUID already has metadata loaded in MongoDB, the metadata
            # should not be loaded to MongoDB again unless replace is True.
            if (kwargs.get("replace") or
                0 == self.database[self.metadata_collection]\
                    .count_documents({"uuid": basket.uuid})
            ):
                self.database[collection].replace_one(
                    {"uuid": basket.uuid},
                    mongo_metadata,
                    upsert=True,
                )

    def load_mongo_manifest(self, uuids: list[str], **kwargs):
        """Load manifest from baskets into the mongo database.

        A manifest.json is created in baskets upon upload. This manifest is
        added to the Mongo database when invoking load_mongo_manifest. UUID,
        and basket_type from the index_table are also added to Mongo for
        referrence back to the datasource.

        Parameters
        ----------
        uuids: [str]
            A list of uuids to add their manifests to the mongo db.
        **collection: str (optional; defaults to self.manifest_collection)
            Manifests will be added to the Mongo collection specified. If not
            provided, populate using self.manifest_collection (which defaults
            to "manifest" if otherwise unspecified).
        """
        collection = kwargs.get("collection", self.manifest_collection)
        if not isinstance(uuids, list):
            uuids = [uuids]
        if not isinstance(uuids[0], str):
            raise TypeError("Invalid datatype for uuids: "
                            "must be a list of strings [str]")
        if not isinstance(collection, str):
            raise TypeError("Invalid datatype for manifest collection: "
                            "must be a string")
        for uuid in uuids:
            basket = self.pantry.get_basket(uuid)
            mongo_manifest = basket.get_manifest()
            if not mongo_manifest:
                continue

            # If the UUID already has the manifest loaded in MongoDB,
            # the manifest should not be loaded to MongoDB again.
            if 0 == self.database[
                collection
            ].count_documents({"uuid": basket.uuid}):
                self.database[collection].insert_one(mongo_manifest)

    def load_mongo_supplement(self, uuids: list[str], **kwargs):
        """Load supplement from baskets into the mongo database.

        A supplement.json is created in baskets upon upload. This supplement
        is added to the Mongo database when invoking load_mongo_supplement.
        UUID, and basket_type from the index_table are also added to Mongo for
        referrence back to the datasource.

        Parameters
        ----------
        uuids: [str]
            A list of uuids to add their supplement to the mongo db.
        **collection: str (optional; defaults to self.supplement_collection)
            Supplements will be added to the Mongo collection specified. If not
            provided, populate using self.supplement_collection (which defaults
            to "supplement" if otherwise unspecified).
        """
        collection = kwargs.get("collection", self.supplement_collection)
        if not isinstance(uuids, list):
            uuids = [uuids]
        if not isinstance(uuids[0], str):
            raise TypeError("Invalid datatype for uuids: "
                            "must be a list of strings [str]")
        if not isinstance(collection, str):
            raise TypeError("Invalid datatype for supplement collection: "
                            "must be a string")

        for uuid in uuids:
            basket = self.pantry.get_basket(uuid)
            supplement = basket.get_supplement()
            if not supplement:
                continue
            mongo_supplement = {}
            mongo_supplement["uuid"] = basket.uuid
            mongo_supplement["basket_type"] = basket.basket_type
            mongo_supplement.update(supplement)

            # If the UUID already has metadata loaded in MongoDB,
            # the metadata should not be loaded to MongoDB again.
            if 0 == self.database[
                collection
            ].count_documents({"uuid": basket.uuid}):
                self.database[collection].insert_one(mongo_supplement)


    def load_mongo(self, uuids: list[str], **kwargs):
        """Load metadata, manifest, and supplement from baskets into the
        mongo database.

        Parameters
        ----------
        uuids: [str]
            A list of uuids to add their data to the mongo db.
        **metadata_collection: str (default=self.metadata_collection)
            Metadata will be added to the Mongo collection specified.
        **manifest_collection: str (default=self.manifest_collection)
            Manifest will be added to the Mongo collection specified.
        **supplement_collection: str (default=self.supplement_collection)
            Supplement will be added to the Mongo collection specified.
        """
        metadata_collection = kwargs.get("metadata_collection",
                                         self.metadata_collection)
        manifest_collection = kwargs.get("manifest_collection",
                                         self.manifest_collection)
        supplement_collection = kwargs.get("supplement_collection",
                                           self.supplement_collection)

        self.load_mongo_metadata(uuids,
                                 collection=metadata_collection,
                                 **kwargs)
        self.load_mongo_manifest(uuids, collection=manifest_collection)
        self.load_mongo_supplement(uuids, collection=supplement_collection)

    def clear_mongo(self, refresh: bool = False):
        """Clear the metadata, manifest, and supplement collections optionally,
        refreshing them from the pantry.

        Parameters
        ----------
        refresh: bool (default=False)
            If True, reload the collections with data retreived from the pantry
        """
        # Drop the referenced database.
        self.database.client.drop_database(self.database_name)

        # Optionally refresh the mongo collections with the current index.
        if refresh:
            self.load_mongo(
                self.pantry.index.to_pandas_df(max_rows=None)['uuid'].to_list()
            )

    def remove_document(self, uuid: str, **kwargs):
        """Delete all documents containing the given uuid from all collections.

        Parameters
        ----------
        uuid: str
            "uuid" will be used to locate and remove the documents from MongoDB
            in the supplement, manifest, and metadata collections.
        **metadata_collection: str (default=self.metadata_collection)
            Metadata will be added to the Mongo collection specified.
        **manifest_collection: str (default=self.manifest_collection)
            Manifest will be added to the Mongo collection specified.
        **supplement_collection: str (default=self.supplement_collection)
            Supplement will be added to the Mongo collection specified.
        """
        metadata_collection = kwargs.get("metadata_collection",
                                         self.metadata_collection)
        manifest_collection = kwargs.get("manifest_collection",
                                         self.manifest_collection)
        supplement_collection = kwargs.get("supplement_collection",
                                           self.supplement_collection)

        collection_names = (metadata_collection,
                            manifest_collection,
                            supplement_collection)

        for collection in collection_names:
            if not isinstance(collection, str):
                raise TypeError("Invalid datatype for collection: "
                                "must be a string")

        for collection in collection_names:
            self.database[collection].delete_many({'uuid':uuid})



"""Wherein is contained the Abstract Base Class for Index."""

import abc
import warnings
from datetime import datetime
from typing import Optional

import pandas as pd
from fsspec import AbstractFileSystem


class IndexABC(abc.ABC):
    """Abstract Base Class for the Index."""

    @abc.abstractmethod
    def __init__(
        self, file_system: AbstractFileSystem, pantry_path: str, **kwargs
    ):
        """Initializes the Index class.

        Parameters
        ----------
        file_system: fsspec object
            The fsspec object which hosts the pantry we desire to index.
        pantry_path: str
            Path to the pantry root which we want to index.
        **setup_config: dict (required)
            Existing metadata for the Index.
        Optional kwargs controlled by concrete implementations.
        """

        self._file_system = file_system
        self._pantry_path = pantry_path
        self.setup_config = kwargs.get("setup_config", {})
        self.generate_config()

    @property
    @abc.abstractmethod
    def file_system(self) -> AbstractFileSystem:
        """The file system of the pantry referenced by this Index."""

    @property
    @abc.abstractmethod
    def pantry_path(self) -> str:
        """The pantry path referenced by this Index."""

    @abc.abstractmethod
    def generate_metadata(self, **kwargs) -> dict:
        """(Deprecated) Generate the metadata for the index."""
        warnings.warn(
            UserWarning("This function is being deprecated in a future weave "
                        "release. Please use generate_config() instead.")
        )
        return self.generate_config(**kwargs)

    @abc.abstractmethod
    def generate_config(self, **kwargs) -> dict:
        """Generate the setup config for the index.

        Parameters
        ----------
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        A dictionary of metadata for the index.
        """
        self.setup_config["index_type"] = str(self)

    @abc.abstractmethod
    def generate_index(self, **kwargs):
        """Populates the index from the file system.

        Generate the index by scraping the pantry and adding the manifest data
        of found baskets to the index.

        Parameters
        ----------
        Optional kwargs controlled by concrete implementations.
        """

    @abc.abstractmethod
    def clear_index(self, refresh: bool = False, **kwargs):
        """Deletes/clears the previously populated index.

        Delete the index entirely, optionally regenerating a new one
        if the refresh flag is set to True.

        Parameters
        ----------
        refresh: bool (default=False)
            Regenerates the index after clearing the old one.
        Optional kwargs controlled by concrete implementations.
        """

    @abc.abstractmethod
    def to_pandas_df(
        self, max_rows: Optional[int] = None, offset: int = 0, **kwargs
    ) -> pd.DataFrame:
        """Returns the pandas dataframe representation of the index.

        Parameters
        ----------
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query

        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        Returns a pandas dataframe of the manifest data of the baskets in the
        pantry.
        """

    @abc.abstractmethod
    def track_basket(self, entry_df: pd.DataFrame, **kwargs):
        """Track a basket (or many baskets) from the pantry with the Index.

        Parameters
        ----------
        entry_df : pandas.DataFrame
            Uploaded baskets to append to the index.
        Optional kwargs controlled by concrete implementations.
        """

    @abc.abstractmethod
    def untrack_basket(self, basket_address: str | list[str], **kwargs):
        """Remove a basket from being tracked of given UUID or path.

        Parameters
        ----------
        basket_address: str or [str]
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket. These may also be passed in
            as a list.
        Optional kwargs controlled by concrete implementations.
        """

    @abc.abstractmethod
    def get_rows(
        self, basket_address: str | list[str], **kwargs
    ) -> pd.DataFrame:
        """Returns a pd.DataFrame row information of given UUID or path.

        Parameters
        ----------
        basket_address: str or [str]
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket. These may also be passed in
            as a list.
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame of the Manifest information for the requested
        basket(s).
        """

    @abc.abstractmethod
    def get_parents(self, basket_address: str, **kwargs) -> pd.DataFrame:
        """Returns a pandas dataframe of all parents of a basket.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing all the manifest data AND generation level
        of parents (and recursively their parents) of the given basket.
        """

    @abc.abstractmethod
    def get_children(self, basket_address: str, **kwargs) -> pd.DataFrame:
        """Returns a pandas dataframe of all children of a basket.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing all the manifest data AND generation level
        of children (and recursively their children) of the given basket.
        """

    @abc.abstractmethod
    def get_baskets_of_type(
        self,
        basket_type: str,
        max_rows: Optional[int] = None,
        offset: int = 0,
        **kwargs,
    ) -> pd.DataFrame:
        """Returns a pandas dataframe containing baskets of basket_type.

        Parameters
        ----------
        basket_type: str
            The basket type to filter for.
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query

        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets of the type.
        """

    @abc.abstractmethod
    def get_baskets_of_label(
        self,
        basket_label: str,
        max_rows: Optional[int] = None,
        offset: int = 0,
        **kwargs,
    ) -> pd.DataFrame:
        """Returns a pandas dataframe containing baskets with label.

        Parameters
        ----------
        basket_label: str
            The label to filter for.
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query

        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets with the
        label.
        """

    @abc.abstractmethod
    def get_baskets_by_upload_time(
        self,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        max_rows: Optional[int] = None,
        offset: int = 0,
        **kwargs,
    ) -> pd.DataFrame:
        """Returns a pandas dataframe of baskets uploaded between two times.

        Parameters
        ----------
        start_time: datetime.datetime (default=None)
            The start datetime object to filter between. If None, will filter
            from the beginning of time.
        end_time: datetime.datetime (default=None)
            The end datetime object to filter between. If None, will filter
            to the current datetime.
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query.

        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets uploaded
        between the start and end times.
        """

        if start_time is not None and not isinstance(start_time,datetime):
            raise ValueError("start_time is not datetime object.")
        if end_time is not None and not isinstance(end_time,datetime):
            raise ValueError("end_time is not datetime object.")

    @abc.abstractmethod
    def query(self, expr: str, **kwargs) -> pd.DataFrame:
        """Returns a pandas dataframe of the results of the query.

        Parameters
        ----------
        expr: str
            An expression passed to the backend. An example could be a SQL or
            pandas query. Largely dependent on concrete implementations.
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame of the resulting query.
        """

    @abc.abstractmethod
    def __len__(self) -> int:
        """Returns the number of baskets in the index."""

    @abc.abstractmethod
    def __str__(self) -> str:
        """Returns the str instantiation type of this Index (ie 'IndexSQL')."""
